/*---------------------------------------------------------------------------
 * Copyright (c) 2024 Arm Limited (or its affiliates). All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *---------------------------------------------------------------------------*/

#include "RTE_Components.h"
#include  CMSIS_device_header

#include "uart_stdout.h"

#include "main.h"

#if defined(ETHOSU_ARCH)
#include "ethosu_driver.h" /* Arm Ethos-U NPU driver header */
#include "ethosu_mem_config.h" /* Arm Ethos-U NPU memory config */

#if defined(ETHOS_U_CACHE_BUF_SZ) && (ETHOS_U_CACHE_BUF_SZ > 0)
static uint8_t cache_arena[ETHOS_U_CACHE_BUF_SZ] CACHE_BUF_ATTRIBUTE;
#else  /* defined (ETHOS_U_CACHE_BUF_SZ) && (ETHOS_U_CACHE_BUF_SZ > 0) */
static uint8_t* cache_arena = NULL;
#endif /* defined (ETHOS_U_CACHE_BUF_SZ) && (ETHOS_U_CACHE_BUF_SZ > 0) */

static uint8_t* get_cache_arena() {
  return cache_arena;
}

static size_t get_cache_arena_size() {
#if defined(ETHOS_U_CACHE_BUF_SZ) && (ETHOS_U_CACHE_BUF_SZ > 0)
  return sizeof(cache_arena);
#else  /* defined (ETHOS_U_CACHE_BUF_SZ) && (ETHOS_U_CACHE_BUF_SZ > 0) */
  return 0;
#endif /* defined (ETHOS_U_CACHE_BUF_SZ) && (ETHOS_U_CACHE_BUF_SZ > 0) */
}

struct ethosu_driver ethosu_drv; /* Default Ethos-U device driver */

/** @brief   Defines the Ethos-U interrupt handler: just a wrapper around the default
 *           implementation. */
static void arm_ethosu_npu_irq_handler(void) {
  /* Call the default interrupt handler from the NPU driver */
  ethosu_irq_handler(&ethosu_drv);
}

/** @brief  Initialises the NPU IRQ */
static void arm_ethosu_npu_irq_init(void) {
#if defined(CORSTONE300_FVP)
  const IRQn_Type ethosu_irqnum = (IRQn_Type)ETHOS_U55_IRQn;
#else
  const IRQn_Type ethosu_irqnum = (IRQn_Type)NPU0_IRQn;
#endif

  /* Register the EthosU IRQ handler in our vector table.
    * Note, this handler comes from the EthosU driver */
  NVIC_SetVector(ethosu_irqnum, (uint32_t)arm_ethosu_npu_irq_handler);

  /* Enable the IRQ */
  NVIC_EnableIRQ(ethosu_irqnum);

  debug("EthosU IRQ#: %u, Handler: 0x%p\n", ethosu_irqnum, arm_ethosu_npu_irq_handler);
}

/** @brief  Initialises the NPU */
static int arm_ethosu_npu_init(void)
{
  int err = 0;

  /* Initialise the IRQ */
  arm_ethosu_npu_irq_init();

  /* Initialise Ethos-U device */
#if defined(CORSTONE300_FVP)
  void* const ethosu_base_address = (void*)(ETHOS_U55_APB_BASE_S);
#else
  void* const ethosu_base_address = (void*)(NPU0_APB_BASE_S);
#endif

  debug("Cache arena: 0x%p\n", get_cache_arena());

  if (0 != (err = ethosu_init(&ethosu_drv,  /* Ethos-U driver device pointer */
                  ethosu_base_address,      /* Ethos-U NPU's base address. */
                  get_cache_arena(),        /* Pointer to fast mem area - NULL for U55. */
                  get_cache_arena_size(),   /* Fast mem region size. */
                  1,                        /* Security enable. */
                  1)))                      /* Privilege enable. */
  {
      printf_err("failed to initialise Ethos-U device\n");
      return err;
  }

  info("Ethos-U device initialised\n");

  /* Get Ethos-U version */
  struct ethosu_driver_version driver_version;
  struct ethosu_hw_info hw_info;

  ethosu_get_driver_version(&driver_version);
  ethosu_get_hw_info(&ethosu_drv, &hw_info);

  info("Ethos-U version info:\n");
  info("\tArch:       v%" PRIu32 ".%" PRIu32 ".%" PRIu32 "\n",
        hw_info.version.arch_major_rev,
        hw_info.version.arch_minor_rev,
        hw_info.version.arch_patch_rev);
  info("\tDriver:     v%" PRIu8 ".%" PRIu8 ".%" PRIu8 "\n",
        driver_version.major,
        driver_version.minor,
        driver_version.patch);
  info("\tMACs/cc:    %" PRIu32 "\n", (uint32_t)(1 << hw_info.cfg.macs_per_cc));
  info("\tCmd stream: v%" PRIu32 "\n", hw_info.cfg.cmd_stream_version);

  return 0;
}

#endif /* if defined(ETHOSU_ARCH) */

int main (void) {
  UartStdOutInit();

#if defined(ETHOSU_ARCH)
  /* Initialise the NPU */
  arm_ethosu_npu_init();
#endif /* defined(ETHOSU_ARCH) */

  return (app_main());
}
